
;==== AN EXAMPLE SCRIPT ====
;This is a script in its external form. Each sentence of a script consists
;of the following components.
; (<keyword> = <subst | nil> <rank> (<pattern1> <transformation rules>)
;                                   (<pattern2> <transformation rules>)
;                                   ... )
;
; (1) keyword  (2) =  (3) replacement of keyword (or nil if no replacement)
; (4) rank (5) lists each of which contains the following:
; (5.1) match pattern  (5.2) pointer to current response (1 for starting)
; (5.3) responses to the input
;
;
;List splicing is done via: (& l)
;Single atom splicing can be done as: ($ l)
;For all patterns, employ the conventions as described in the MATCH function.
;
;Each such script will be read by HONEY and converted into an indexed
;structure for more efficient retrieval of keywords.
;
;Date of last modification: December 12, 1986.
;

(defun talk (word)
  (member word '(said says say talk speak spoke uttered goes thinks think
  believe believes)))

(defun family (word)
  (member word '(mother father dad mom mama papa brother sister cousin
                 uncle aunt yia-yia grandpa grandma oma opa)))

(defun insult (word)
  (member word '(stupid silly brain-damaged)))

(defun names (word)
  (member word '(jerk idiot ass asshole pig dog)))

(defun places (word)
  (member word '(work office coffee lunch dinner break mall shop)))

(defun move (word)
  (member word '(went drove walked rode)))

(defun nots (word)
  (member word '(didn't not don't won't shan't)))


(setq *script* '(

(Where have you been for so long?)

(*remark* = nil 0 (nil 1 (I know you have been seeing another woman.)
                         (Can you explain one more time where you were?)
                         (Alright, tell me all about your previous life.)
                         (I hate it when you leave me alone for a long time.)
                         (You know, you used to treat me better.)
                         (you are giving me a headache.)
                         (you are so confusing today.)
                         (oh, that i don't forget, did you take out the
                          garbage?)
                         (Elaborate, please!))

)

(*memory* = nil 0 ((my (restrict (? l) family) (restrict ? talk) you are (+ k)) 1
                      (by the way, what did your ($ l) mean by saying i'm (& k)?))
                  ((+ mother *) 1
                      (Before you said something about your mother.)
                      (Why did you mention your mother before?))
)

(sorry = nil 4 ((I am +) 1 (Now you are sorry.)
                      (Think first, then apologize.)
                      (If you would think before you speak, you wouldn't
                       have to apologize that much.)
                      (This time I'll accept your apology.))
               ((i'm +) 1 (Alright you are sorry, but keep on talking.))
               ((+) 1 (don't grovel, alright?)
                      (think first next time.)
                      (i hate it when you do something and then apologize.))
)

(no = nil 1 ((+) 1 (why not?) (come on...)
                   (please?) (uhh, rather firm.))

)

(yes = nil 1 ((+) 1 (A little more detail, please.)
                    (Yes?)
                    (My godness, aren't we brief today.)
                    (elucidate, will you?))

)

(why = nil 1 ((why) 1 (why not?)
                      (i'm curious by nature, you know.)
                      (i need to know more details))

)

(where = nil 1 ((where) 1 (i don't know.))
               ((where have you (+ l)) 1
                       (what's it to you where i've (& l) ?))
               ((where are (+ l)) 1
                       (i don't care where (& l) are.)
                       (don't ask me.)
                       (what did you do with (& l) ?))

)

(my = your 4 ( (my (restrict (? k) family) (restrict ? talk) that you are (+ l)) 1
                     (what does your ($ k) mean by: i'm (& l) ?)
                     (how come your ($ k) would say that i'm (& l) ?)
                     (what was the last thing after: your ($ k) said i'm ?) )
             ( ((restrict (? k) family) (restrict ? talk) (+ l)) 1
                     (your ($ k) is is very strange.) )

)

(how = nil 1 ((+) 1 (= why))
)

(yourself = myself 5 ((+) 1 (what is wrong with me?)
                            (why are you picking on me?))

)

(don't = nil 5 ((i don't (+ l)) 1
                    (what? you don't (& l) ?)
                    (why don't you (& l)))
               ((you don't (+ l)) 1
                    (i very well (& l))
                    (yes, i do (+ l)))

)

(exactly = nil 1 ((+) 1 (= yes))
)

(your = my 1 ((your (restrict (? l) family) +) 1 (don't pick on my ($ l)) )
)

(course = nil 1 ((of course +) 1
                   (how come you're so sure?)
                   (what do you mean with 'of course'?))

)

(leave = nil 3 ((leave me alone) 1
                   (you never leave me alone either.)
                   (alright, if you want to be that way.)
                   (are you upset?)
                   (you don't like me anymore, is that it?))

)

(nothing = nil 3 ((nothing) 1
                   (talk to me, will you.)
                   (what do you mean by nothing?))
             ((+ nothing) 1
                   (don't tell me it's nothing!))

)

(father = nil 8 ( (my father (restrict ? talk) (+ l)) 1
                       (how can he say (& l)) )
                ( (+) 1
                       (= family) )
)

(mother = nil 8 ( (+) 1 (= family))
)

(brother = nil 8 ( (+) 1 (= family))
)

(sister = nil 8 ( (+) 1 (= family))
)

(family = nil 0 (nil 1 (i think you like your family more than me.)
                       (not again your family.)
                       (you never do anything with me. Always your family.)
                       (i never liked your relatives anyway.))

)

(because = nil 1 ((because) 1
                     (because what?))
                 ((because (+ l)) 1
                     (why do (& l)))
)

(you = me 3 ((you are (restrict (? l) insult)) 1
                   (you are ($ l) yourself.)
                   (you have no right to call me ($ l)))
            ((you are a (restrict (? l) names)) 1
                   (you are a ($ l) yourself!)
                   (how dare you call me a ($ l)))
            ((? (restrict (? k) family) (restrict ? talk) you are (+ l)) 1
                   (how would your ($ k) know that i'm (& l)))
)

(thank = nil 4 ((thank +) 1 (you're welcome.))
)

(at = nil 1 ((+ (restrict (? l) places) +) 1
                          (you never took me to ($ l)))

)

(i = you 4 ((i have been (+ l)) 1
               (why were you (& l)))
           ((i (restrict ? move) to (+ l)) 1
               (what did you do at (& l)))
           ((i ? with (? l)) 1
               (who is ($ l) ?))
           ((i (restrict (? k) nots) (+ l)) 1
               (why ($ k) you (& l) ?)
               (how come your ($ k) (& l) ?))
           ((i +) 1
               (stop worrying about yourself. Think about me!))
)



))
;;; End of script.